@@ -2,7 +2,8 @@ import copy
 import json
 import warnings
 from abc import ABC
-from collections.abc import AsyncGenerator
+from collections.abc import AsyncGenerator, Callable
+from http import HTTPStatus
 from typing import Any, Generic, Literal, TypeVar, Union, cast, overload
 
 import aiohttp
@@ -29,10 +30,12 @@ class AsyncClient(AbstractClient, ABC):
         authorization: Union[str, None] = None,
         extra_headers: Union[dict, None] = None,
         aiohttp_config: Union[dict, None] = None,
+        *,
+        dump_resource: Callable[[Any], dict] = lambda x: dict(x),
     ):
         self.aiohttp_config = aiohttp_config or {}
 
-        super().__init__(url, authorization, extra_headers)
+        super().__init__(url, authorization, extra_headers, dump_resource=dump_resource)
 
     async def execute(
         self,
@@ -43,27 +46,31 @@ class AsyncClient(AbstractClient, ABC):
     ):
         return await self._do_request(method, path, data=data, params=params)
 
-    @overload
-    async def get(self, resource_type_or_resource: TResource, id: None = None) -> TResource:
-        ...
-
     @overload
     async def get(
-        self, resource_type_or_resource: type[TResource], id: Union[str, None] = None
+        self, resource_type_or_resource_or_ref: TResource, id_or_ref: None = None
     ) -> TResource:
         ...
 
     @overload
-    async def get(self, resource_type_or_resource: str, id: Union[str, None] = None) -> Any:
+    async def get(
+        self, resource_type_or_resource_or_ref: type[TResource], id_or_ref: Union[str, None] = None
+    ) -> TResource:
+        ...
+
+    @overload
+    async def get(
+        self, resource_type_or_resource_or_ref: str, id_or_ref: Union[str, None] = None
+    ) -> dict:
         ...
 
     async def get(
         self,
-        resource_type_or_resource: Union[str, type[TResource], TResource],
-        id: Union[str, None] = None,  # noqa: A002
-    ) -> Union[TResource, Any]:
+        resource_type_or_resource_or_ref: Union[str, type[TResource], TResource],
+        id_or_ref: Union[str, None] = None,
+    ) -> Union[TResource, dict]:
         resource_type, resource_id, custom_resource_class = get_resource_type_id_and_class(
-            resource_type_or_resource, id
+            resource_type_or_resource_or_ref, id_or_ref
         )
 
         if resource_id is None:
@@ -95,7 +102,7 @@ class AsyncClient(AbstractClient, ABC):
         *,
         _search_params: Union[dict, None] = None,
         _as_dict: Literal[True] = True,
-    ) -> Any:
+    ) -> dict:
         ...
 
     async def save(
@@ -108,8 +115,8 @@ class AsyncClient(AbstractClient, ABC):
         _search_params: Union[dict, None] = None,
         # _as_dict is a private api used internally
         _as_dict: bool = False,
-    ) -> Union[TResource, Any]:
-        data = serialize(resource)
+    ) -> Union[TResource, dict]:
+        data = serialize(self.dump_resource(resource), drop_nulls_from_dicts=fields is None)
         if fields:
             if not resource.id:
                 raise TypeError("Resource `id` is required for update operation")
@@ -137,37 +144,42 @@ class AsyncClient(AbstractClient, ABC):
 
     @overload
     async def patch(
-        self, resource_type_or_resource: TResource, id: None = None, **kwargs
+        self, resource_type_or_resource_or_ref: TResource, id_or_ref: None = None, **kwargs
     ) -> TResource:
         ...
 
     @overload
     async def patch(
-        self, resource_type_or_resource: type[TResource], id: Union[str, None] = None, **kwargs
+        self,
+        resource_type_or_resource_or_ref: type[TResource],
+        id_or_ref: Union[str, None] = None,
+        **kwargs,
     ) -> TResource:
         ...
 
     @overload
     async def patch(
-        self, resource_type_or_resource: str, id: Union[str, None] = None, **kwargs
-    ) -> Any:
+        self, resource_type_or_resource_or_ref: str, id_or_ref: Union[str, None] = None, **kwargs
+    ) -> dict:
         ...
 
     async def patch(
         self,
-        resource_type_or_resource: Union[str, type[TResource], TResource],
-        id: Union[str, None] = None,  # noqa: A002
+        resource_type_or_resource_or_ref: Union[str, type[TResource], TResource],
+        id_or_ref: Union[str, None] = None,
         **kwargs,
-    ) -> Union[TResource, Any]:
+    ) -> Union[TResource, dict]:
         resource_type, resource_id, custom_resource_class = get_resource_type_id_and_class(
-            resource_type_or_resource, id
+            resource_type_or_resource_or_ref, id_or_ref
         )
 
         if resource_id is None:
             raise TypeError("Resource `id` is required for patch operation")
 
         response_data = await self._do_request(
-            "patch", f"{resource_type}/{resource_id}", data=serialize(kwargs)
+            "patch",
+            f"{resource_type}/{resource_id}",
+            data=serialize(kwargs, drop_nulls_from_dicts=False),
         )
 
         if custom_resource_class:
@@ -177,11 +189,11 @@ class AsyncClient(AbstractClient, ABC):
 
     async def delete(
         self,
-        resource_type_or_resource: Union[str, type[TResource], TResource],
-        id: Union[str, None] = None,  # noqa: A002
-    ):
+        resource_type_or_resource_or_ref: Union[str, type[TResource], TResource],
+        id_or_ref: Union[str, None] = None,
+    ) -> Any:
         resource_type, resource_id, _ = get_resource_type_id_and_class(
-            resource_type_or_resource, id
+            resource_type_or_resource_or_ref, id_or_ref
         )
 
         if resource_id is None:
@@ -196,6 +208,8 @@ class AsyncClient(AbstractClient, ABC):
         path: str,
         data: Union[dict, None] = None,
         params: Union[dict, None] = None,
+        extra_headers: Union[dict, None] = None,
+        *,
         returning_status: Literal[False] = False,
     ) -> Any:
         ...
@@ -207,19 +221,26 @@ class AsyncClient(AbstractClient, ABC):
         path: str,
         data: Union[dict, None] = None,
         params: Union[dict, None] = None,
+        extra_headers: Union[dict, None] = None,
+        *,
         returning_status: Literal[True] = True,
     ) -> tuple[Any, int]:
         ...
 
-    async def _do_request(
+    async def _do_request(  # noqa: PLR0913
         self,
         method: str,
         path: str,
         data: Union[dict, None] = None,
         params: Union[dict, None] = None,
+        extra_headers: Union[dict, None] = None,
+        *,
         returning_status=False,
     ) -> Union[Any, tuple[Any, int]]:
         headers = self._build_request_headers()
+        if extra_headers:
+            headers = {**headers, **extra_headers}
+
         url = self._build_request_url(path, params)
         async with aiohttp.ClientSession(headers=headers) as session:
             async with session.request(method, url, json=data, **self.aiohttp_config) as r:
@@ -228,6 +249,9 @@ class AsyncClient(AbstractClient, ABC):
                     r_data = json.loads(raw_data, object_hook=AttrDict) if raw_data else None
                     return (r_data, r.status) if returning_status else r_data
 
+                if r.status == 304:  # noqa: PLR2004
+                    return (None, r.status) if returning_status else None
+
                 if r.status in (404, 410):
                     raise ResourceNotFound(await r.text())
 
@@ -284,6 +308,8 @@ class AsyncResource(
         return cast(TResource, self)
 
     async def patch(self, **kwargs) -> TResource:
+        if not self.id:
+            raise TypeError("Resource `id` is required for patch operation")
         super(BaseResource, self).update(**kwargs)
         await self.save(fields=list(kwargs.keys()))
 
@@ -292,12 +318,19 @@ class AsyncResource(
     async def delete(self):
         if not self.id:
             raise TypeError("Resource `id` is required for delete operation")
-        return await self.__client__.delete(self)
+        return await self.__client__.delete(self.reference)
 
     async def refresh(self) -> TResource:
-        data = await self.__client__._do_request("get", self._get_path())
-        super(BaseResource, self).clear()
-        super(BaseResource, self).update(**data)
+        data, status = await self.__client__._do_request(
+            "get",
+            self._get_path(),
+            extra_headers={"If-None-Match": self["meta"]["versionId"]},
+            returning_status=True,
+        )
+
+        if status != HTTPStatus.NOT_MODIFIED:
+            super(BaseResource, self).clear()
+            super(BaseResource, self).update(**data)
 
         return cast(TResource, self)
 
@@ -394,8 +427,8 @@ class AsyncSearchSet(
         data = await self.client._fetch_resource(self.resource_type, self.params)
         data_resource_type = data.get("resourceType", None)
 
-        if data_resource_type == "Bundle":
-            for item in data["entry"]:
+        if data_resource_type == "Bundle" and not self.custom_resource_class:
+            for item in data.get("entry", []):
                 item.resource = self._dict_to_resource(item.resource)
 
         return data
@@ -438,7 +471,7 @@ class AsyncSearchSet(
         response_data, status_code = await self.client._do_request(
             "post",
             self.resource_type,
-            serialize(resource),
+            serialize(self.client.dump_resource(resource)),
             self.params,
             returning_status=True,
         )
@@ -451,7 +484,7 @@ class AsyncSearchSet(
         response_data, status_code = await self.client._do_request(
             "put",
             self.resource_type,
-            serialize(resource),
+            serialize(self.client.dump_resource(resource)),
             self.params,
             returning_status=True,
         )
@@ -465,13 +498,16 @@ class AsyncSearchSet(
             DeprecationWarning,
             stacklevel=2,
         )
-        data = serialize(_resource if _resource is not None else kwargs)
+        data = serialize(
+            self.client.dump_resource(_resource) if _resource is not None else kwargs,
+            drop_nulls_from_dicts=False,
+        )
         response_data = await self.client._do_request(
             "PATCH", self.resource_type, data, self.params
         )
         return self._dict_to_resource(response_data)
 
-    async def delete(self) -> Any:
+    async def delete(self) -> tuple[Any, int]:
         return await self.client._do_request(
             "delete", self.resource_type, params=self.params, returning_status=True
         )
